;       Filename:       cpuid.asm
;
;       This program has been developed by Intel Corporation.  You have
;       Intel's permission to incorporate this source code into your
;       product royalty free.
;
;       Intel specifically disclaims all warranties, express or implied,
;       and all liability, including consequential and other indirect
;       damages, for the use of this code, including liability for
;       infringement of any proprietary rights.  Intel does not assume
;       any responsibility for any errors which may appear in this code
;       nor any responsibility to update it.
;
;       If this code is assembled with MASM with no options specified, it
;       runs correctly on an 8086/8088, 80286, 80386, 80486, and 
;       Pentium(tm) processor.
;
;       Modified by Phil Randal to check for Pentium FPU bug
;

	TITLE   CPUID
	.model  small
	.stack  100h
	.8086

TRUE            equ     1
FAMILY_MASK     equ     0f00h
FAMILY_SHIFT    equ     8
MODEL_MASK      equ     0f0h
MODEL_SHIFT     equ     4
STEPPING_MASK   equ     0fh
FPU_FLAG        equ     1h
MCE_FLAG        equ     80h
CMPXCHG8B_FLAG  equ     100h

	.data
id_msg          db 0Dh,0Ah
		db 'This system has a$'
unknown_proc    db 'n unknown processor$'
c8086           db 'n 8086/8088 processor$'
c286            db 'n 80286 processor$'
c386            db 'n 80386 processor$'
c486            db 'n 80486DX, 80486DX2 processor or 80487SX math coprocessor.$'
c486nfp         db 'n 80486SX processor.$'
fp_8087         db ' and an 8087 math coprocessor.$'
fp_80287        db ' and an 80287 math coprocessor.$'
fp_80387        db ' and an 80387 math coprocessor.$'
Intel486_msg    db ' Genuine Intel486(TM) processor.$'
Genuine_486DX   db ' Genuine Intel486(TM) DX processor.$'
Genuine_486SX   db ' Genuine Intel486(TM) SX processor.$'
Genuine_DX2     db ' Genuine IntelDX2(TM) processor.$'
Genuine_SX2     db ' Genuine IntelSX2(TM) processor.$'
Genuine_DX4     db ' Genuine IntelDX4(TM) processor.$'
Genuine_DX2A    db ' Genuine IntelDX2(TM) processor.$'
Genuine_Pentium db ' Genuine Intel Pentium(TM) processor.$'
Pentium_OD_msg  db ' Genuine Intel Pentium(TM) OverDrive(TM) Processor.$'
Genuine_Intel   db ' Genuine Intel processor.$'
copyright       db 0Dh,0Ah
		db 'Copyright 1993,1994 Intel Corporation'
		db 0Dh,0Ah
		db 'Pentium bug detection code by Phil Randal$'
msg_table       dw offset Genuine_486DX
		dw offset Genuine_486DX
		dw offset Genuine_486SX
		dw offset Genuine_DX2
		dw offset Intel486_msg
		dw offset Genuine_SX2
		dw offset Intel486_msg
		dw offset Genuine_DX2A
		dw offset Genuine_DX4
		dw offset Intel486_msg
		dw offset Intel486_msg
		dw offset Intel486_msg
		dw offset Intel486_msg
		dw offset Intel486_msg
		dw offset Intel486_msg
		dw offset Intel486_msg
crlf            db 0Dh,0Ah,'$'
OverDrive       db 0Dh,0Ah
		db 'The processor is an OverDrive(TM) upgrade processor$'
Upgrade         db 0Dh,0Ah
		db 'The processor is the upgrade processor in a dual processor system$'
NonIntel        db ' processor which is not a Genuine Intel processor',0Dh,0Ah
		db 'and as a result, the CPUID detection information',0Dh,0Ah
		db 'detection information cannot be determined.$'
stepping        db 0
cpu_model       db 0
cpu_type        db 0
fpu_type        db 0
cpuid_flag      db 0
intel_proc      db 0
vendor_id       db '------------'
intel_id        db 'GenuineIntel'
s_eax           dd 0
s_ecx           dd 0
feature_bits    dd 0
s_ebx           dd 0
fp_status       dw 0
bugnr           dt 824633702441.0
diff            dt 0.0
div_bug         db 0
div_bug_msg     db 0Dh,0Ah
		db 'This processor has the floating point division bug:'
                db 0Dh,0Ah
                db '(1.0/824633702441.0)*824633702441.0 is not equal to 1.0!$'

;
;       The purpose of this code is to identify the processor and
;       coprocessor that is currently in the system.  The program first
;       determines the processor id.  When that is accomplished,
;       the program then determines whether a coprocessor
;       exists in the system.  If a coprocessor or integrated
;       coprocessor exists, the program identifies
;       the coprocessor id.  The program then prints out
;       the CPU and floating point presence and type.
;
	.code
start:  mov     ax,@data
	mov     ds,ax          ; set segment register
	mov     es,ax          ; set segment register
	and     sp,0FFFCh
	call    get_cpuid
	call    get_fpuid
	call    print
	mov     ax,4c00h       ; terminate program
	int     21h


get_cpuid proc
;
;       This procedure determines the type of CPU in a system
;       and sets the cpu_type variable with the appropriate
;       value.
;       All registers are used by this procedure, none are preserved.

;       Intel 8086 CPU check
;       Bits 12-15 of the FLAGS register are always set on the
;       8086 processor.
;
check_8086:
	mov     ax,ds
	mov     es,ax
	pushf                   ; push original FLAGS
	pop     ax              ; get original FLAGS
	mov     cx,ax           ; save original FLAGS
	and     ax,0fffh        ; clear bits 12-15 in FLAGS
	push    ax              ; save new FLAGS value on stack
	popf                    ; replace current FLAGS value
	pushf                   ; get new FLAGS
	pop     ax              ; store new FLAGS in AX
	and     ax,0f000h       ; if bits 12-15 are set, then CPU
	cmp     ax,0f000h       ;   is an 8086/8088
	mov     cpu_type,0      ; turn on 8086/8088 flag
	jne     check_80286     ; jump if CPU is 8086/8088
	jmp     end_get_cpuid   ; jump if CPU is 8086/8088

;       Intel 286 CPU check
;       Bits 12-15 of the FLAGS register are always clear on the
;       Intel 286 processor in real-address mode.
;
check_80286:
	or      cx,0f000h       ; try to set bits 12-15
	push    cx              ; save new FLAGS value on stack
	popf                    ; replace current FLAGS value
	pushf                   ; get new FLAGS
	pop     ax              ; store new FLAGS in AX
	and     ax,0f000h       ; if bits 12-15 clear, CPU=80286
	mov     cpu_type,2      ; turn on 80286 flag
	jne     check_80386
	jmp     end_get_cpuid   ; if no bits set, CPU is 80286

;       Intel386 CPU check
;       The AC bit, bit #18, is a new bit introduced in the EFLAGS
;       register on the Intel486 DX CPU to generate alignment faults.
;       This bit cannot be set on the Intel386 CPU.
;
check_80386:
;       It is now safe to use 32-bit opcode/operands
	.386
	pushfd                  ; push original EFLAGS
	pop     eax             ; get original EFLAGS
	mov     ecx,eax         ; save original EFLAGS
	xor     eax,40000h      ; flip AC bit in EFLAGS
	push    eax             ; save new EFLAGS value on stack
	popfd                   ; replace current EFLAGS value
	pushfd                  ; get new EFLAGS
	pop     eax             ; store new EFLAGS in EAX
	xor     eax,ecx         ; can't toggle AC bit, CPU=80386
	mov     cpu_type,3      ; turn on 80386 CPU flag
	jz      end_get_cpuid   ; jump if 80386 CPU
	push    ecx
	popfd                   ; restore AC bit in EFLAGS first

;       Intel486 DX CPU, Intel487 SX NDP, and Intel486 SX CPU check
;       Checking for ability to set/clear ID flag (Bit 21) in EFLAGS
;       which indicates the presence of a processor
;       with the ability to use the CPUID instruction.
;
check_80486:
	mov     cpu_type,4      ; turn on 80486 CPU flag
	mov     eax,ecx         ; get original EFLAGS
	xor     eax,200000h     ; flip ID bit in EFLAGS
	push    eax             ; save new EFLAGS value on stack
	popfd                   ; replace current EFLAGS value
	pushfd                  ; get new EFLAGS
	pop     eax             ; store new EFLAGS in EAX
	xor     eax,ecx         ; can't toggle ID bit,
	je      end_get_cpuid   ;   CPU=80486

;       Execute CPUID instruction to determine vendor, family,
;       model and stepping.
;
check_vendor:
	.586
	mov     cpuid_flag,1             ; set flag indicating use of CPUID inst.
	push    ebx
	push    esi
	push    edi
	xor     eax,eax                  ; set up input for CPUID instruction
	cpuid
	mov     dword ptr vendor_id,ebx  ; setup to test for vendor id
	mov     dword ptr vendor_id+4,edx
	mov     dword ptr vendor_id+8,ecx
	mov     si,offset vendor_id
	mov     di,offset intel_id
	mov     cx,length intel_id
compare:
	cld
	rep     cmpsb                   ; compare vendor id to "GenuineIntel"
	jnz     cpuid_data              ; if not zero, not an Intel CPU,

intel_processor:
	mov     intel_proc,1

cpuid_data:
	test    eax,eax
	jz      pop_regs                ; Early Pentium
	mov     eax,1
	cpuid
	mov     s_eax,eax
	mov     s_ebx,ebx
	mov     feature_bits,edx
	mov     s_ecx,ecx
	and     ah,0Fh
	mov     cpu_type,ah
	mov     bl,ah
        xor     bh,bh
        mov     ah,al
        shr     ah,4
        and     al,0Fh
        mov     stepping,al
        mov     cpu_model,ah
        cmp     cpu_type,5
	jne     pop_regs

; Check Pentium Divide Bug

	call    testndp
	je      pop_regs
	mov     div_bug,1

pop_regs:
	pop     edi
	pop     esi
	pop     ebx

end_get_cpuid:
	ret
get_cpuid endp

testndp proc
; Check for FPU bug - returns Z/NZ for OK/Bad
        finit
	fld1
	fld     bugnr
	fdivp   st(1),st
	fld     bugnr
	fmulp   st(1),st
	fld1
	fsubp   st(1),st
        fstp    diff
        fwait
        mov     ax,word ptr diff+6    ; check lower 62 bits of mantissa
        and     ah,3fh                ; throw away two MS bits
        or      ax,word ptr diff+4
        or      ax,word ptr diff+2
        or      ax,word ptr diff
        ret
testndp endp

;******************************************************************

get_fpuid proc
;
;       This procedure determines the type of FPU in a system
;       and sets the fpu_type variable with the appropriate
;       value.
;       All registers are used by this procedure, none are preserved.

;       Coprocessor check
;       The algorithm is to determine whether the floating-point
;       status and control words can be written to.  If not, no
;       coprocessor exists.  If the status and control words can be
;       written to, the correct coprocessor is then determined
;       depending on the processor id.  The Intel386 CPU can
;       work with either an Intel287 NDP or an Intel387 NDP.
;       The infinity of the coprocessor must be
;       checked to determine the correct coprocessor id.

	fninit                  ; reset FP status word
	mov     fp_status,5a5ah ; initialize temp word to
				; non-zero value
	fnstsw  fp_status       ; save FP status word
	mov     ax,fp_status    ; check FP status word
	cmp     al,0            ; see if correct status with
				; written
	mov     fpu_type,0      ; no fpu present
	jne     end_get_fpuid

check_control_word:
	fnstcw  fp_status       ; save FP control word
	mov     ax,fp_status    ; check FP control word
	and     ax,103fh        ; see if selected parts
				; looks OK
	cmp     ax,3fh          ; check that 1's & 0's
				; correctly read
	mov     fpu_type,0      ; No FPU
	jne     end_get_fpuid
	mov     fpu_type,1      ; Assume 8087 for the moment

;
;   80287/80387 check for the Intel386 CPU
;
check_infinity:
	cmp     cpu_type,3
	jne     end_get_fpuid
	wait    
	fld1                    ; must use default control from FNINIT
	wait    
	fldz                    ; form infinity
	wait    
	fdivp   st(1),st        ; 8087 and Intel287 NDP say +inf = -inf
	wait    
	fld     st
	wait    
	fchs                    ; Intel387 NDP says +inf <> -inf
	wait
	fcompp  st(1),st        ; see if they are the same and remove them
	fstsw   fp_status       ; look at status from FCOMPP
	mov     ax,fp_status
	mov     fpu_type,2      ; store Intel287 NDP for fpu type
	sahf                    ; see if infinities matched
	jz      end_get_fpuid   ; jump if 8087 or Intel287 is present
	mov     fpu_type,3      ; store Intel387 NDP for fpu type
end_get_fpuid:
	ret
get_fpuid endp


;*********************************************************************
                
                .8086
print           proc near
		mov     dx,offset id_msg
		mov     ah,9
		int     21h
			   
		cmp     cpuid_flag,1
		jnz     print_86
		jmp     cmp_vendor
;---------------------------------------------------------

print_86:
		cmp     cpu_type,0
		jnz     print_286
		mov     dx,offset c8086
		mov     ah,9
		int     21h
			   
		cmp     fpu_type,0
		jnz     print_87
		jmp     print_copyright
;---------------------------------------------------------

print_87:
		mov     dx,offset fp_8087
		mov     ah,9
		int     21h
			   
		jmp     print_copyright
;---------------------------------------------------------

print_286:
		cmp     cpu_type,2
		jnz     print_386
		mov     dx,offset c286
		mov     ah,9
		int     21h
			   
		cmp     fpu_type,0
		jnz     print_287
		jmp     print_copyright
;---------------------------------------------------------

print_287:
		mov     dx,offset fp_80287
		mov     ah,9
		int     21h   
			      
		jmp     print_copyright
;---------------------------------------------------------

print_386:
		cmp     cpu_type,3
		jnz     print_Intel486sx
		mov     dx,offset c386
		mov     ah,9
		int     21h
			   
		cmp     fpu_type,0
		jnz     print_387
		jmp     print_copyright
;---------------------------------------------------------

print_387:
		cmp     fpu_type,2
		jz      print_287
		mov     dx,offset fp_80387
		mov     ah,9
		int     21h
			   
		jmp     print_copyright
;---------------------------------------------------------

print_Intel486sx:
		cmp     cpu_type,4
		jnz     print_unknown
		mov     dx,offset c486nfp
		cmp     fpu_type,0
		jz      print_486
		mov     dx,offset c486

print_486:
		mov     ah,9
		int     21h
			   
		jmp     print_copyright
;---------------------------------------------------------

print_unknown:
		mov     dx,offset unknown_proc
		jmp     print_486
;---------------------------------------------------------

cmp_vendor:
		cmp     intel_proc,1
		jnz     print_Non_Intel
		cmp     cpu_type,4
		jnz     check_Pentium
		.486
                mov     ax,word ptr s_eax
		shr     ax,4
                and     eax,0Fh
		mov     dx,msg_table[eax*2]
		.8086
                jmp     print_str
;---------------------------------------------------------

check_Pentium:
		cmp     cpu_type,5
		jnz     print_genuine
		mov     ax,word ptr s_eax
		test    ax,1000h
		jz      print_Pentium_2
		mov     dx,offset Pentium_OD_msg
		jmp     print_str
;---------------------------------------------------------

print_Pentium_2:
		mov     dx,offset Genuine_Pentium
		jmp     print_str
;---------------------------------------------------------

print_genuine:
		mov     dx,offset Genuine_Intel

print_str:
		mov     ah,9
		int     21h
			   
		test    ax,2000h
		jz      print_copyright
		mov     dx,offset Upgrade
		mov     ah,9
		int     21h

		jmp     print_copyright
;---------------------------------------------------------

print_Non_Intel:
		mov     dx,offset NonIntel
		mov     ah,9
		int     21h

print_copyright:
		cmp     div_bug,1
		jne     really_print_copyright
		mov     dx,offset div_bug_msg
		mov     ah,9
		int     21h

really_print_copyright:
		mov     dx,offset crlf
		mov     ah,9
		int     21h
			   
		mov     dx,offset copyright
		mov     ah,9
		int     21h
			   
		mov     dx,offset crlf
		mov     ah,9
		int     21h
			   
		ret     
print           endp

	end     start
